// Copyright 2019-2024 Yuji Tanaka
// HP : http://galaxystar.image.coocan.jp/
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
/*
 * ESP32S3 MP3 player for I2S and ILI9163
 *   Please install ESP8266Audio in libraries.
 *   
 *   SD card : "*.wav", "*.mp3"
 *
 * 2024'11/10 Yuji Tanaka 
 */
#include <Arduino.h>

#include <WiFi.h>
#include "FS.h"
#include "SD.h"
#include "SPI.h"
#include <driver/i2s.h>
#include <ILI9163_buf.h>

#include "AudioFileSourceSD.h"
#include "AudioFileSourceID3.h"
#include "AudioGeneratorMP3.h"
#include "AudioGeneratorWAV.h"
#include "AudioOutputI2S.h"

AudioGeneratorMP3 *mp3;
AudioGeneratorWAV *wav;
AudioFileSourceSD *file;
AudioOutputI2S *out;
AudioFileSourceID3 *id3;

#define TFT_DC  48
#define TFT_CS  47
#define TFT_RST 45
#define SD_CS   10

ILI9163_buf tft = ILI9163_buf(TFT_CS, TFT_DC, TFT_RST);

#define MAX_LINE 16
int _width, _height;

#define MAX_FILE 128
File root;
char fname[MAX_FILE][64];
char ftype[MAX_FILE];
int fnum;

int  num, old_num;
int  disp_num;

char old_fname[32];
int  fmode;
int  I2S_stop;

int StickXPin = 6;
int StickYPin = 7;
int StickX;
int StickY;

int ButtonAPin = 15;
int ButtonBPin = 16;
int ButtonA;
int ButtonB;
int OldA;
int OldB;

TaskHandle_t th;

// Called when a metadata event occurs (i.e. an ID3 tag, an ICY block, etc.
void MDCallback(void *cbData, const char *type, bool isUnicode, const char *string)
{
  (void)cbData;
  Serial.printf("ID3 callback for: %s = '", type);

  if (isUnicode) {
    string += 2;
  }
  
  while (*string) {
    char a = *(string++);
    if (isUnicode) {
      string++;
    }
    Serial.printf("%c", a);
  }
  Serial.printf("'\n");
  Serial.flush();
}

void setup()
{
  WiFi.mode(WIFI_OFF); 
  Serial.begin(115200);
  
  // LCD
  tft.begin(40000000);
  tft.setRotation(2);
  _width=tft.width();
  _height=tft.height();
  
  tft.fillRect(0, 0, _width, _height, ILI9163_BLACK);
  tft.setTextSize(1);
  tft.setTextColor(ILI9163_WHITE, ILI9163_BLACK);
  tft.setCursor(_width/2-64, _height/2);
  tft.print("Now read SD card");
  tft.setCursor(16, _height/2+32);
  tft.print("WAV and MP3");
  tft.display();

  Serial.print("Initializing SD card...");
  if (!SD.begin()) {
    show_no_sd();
    Serial.println("SD failed!");
    while(1);
  }
  root = SD.open("/");
  strcpy(old_fname, "/");

  fnum=0;
  strcpy(old_fname, "/");
  printDirectory(root, 0, ".wav", ".mp3");

  Serial.println("OK!");
  
  pinMode(ButtonAPin, INPUT_PULLUP);
  pinMode(ButtonBPin, INPUT_PULLUP);
  
  audioLogger = &Serial;

  out = new AudioOutputI2S(I2S_NUM_0, 0, 6, false);
  wav = new AudioGeneratorWAV();
  mp3 = new AudioGeneratorMP3();

  num=0;
  old_num=-1;
  fmode=0;

  xTaskCreatePinnedToCore(main_loop, "main_loop", 8192, NULL, 3, &th, 0);
}

void loop()
{
  // Idle
  if (fmode==0){
    fmode=0;
  // File and Audio set
  }else if (fmode==1){
    file = new AudioFileSourceSD(fname[num]);
    id3 = new AudioFileSourceID3(file);
    id3->RegisterMetadataCB(MDCallback, (void*)"ID3TAG");
    Serial.println(fname[num]);
    Serial.println(file->getSize());

    if (ftype[num]==0){
      wav->begin(id3, out);
      fmode=2;
    }else if (ftype[num]==1){
      mp3->begin(id3, out);
      fmode=2;
    }
    show_play();
  // Play    
  }else{
    if (ftype[num]==0 && wav->isRunning()) {
      if (!wav->loop() || I2S_stop==1){
        wav->stop();
        if (I2S_stop==1){
          fmode=0;
          I2S_stop=0;
          old_num=-1;
        }else{
          fmode=1;
          num++;
          if (num>=fnum)
            num=0;
        }
      }
    }else if (ftype[num]==1 && mp3->isRunning()) {
      if (!mp3->loop() || I2S_stop==1){
        mp3->stop();
        if (I2S_stop==1){
          fmode=0;
          I2S_stop=0;
          old_num=-1;
        }else{
          fmode=1;
          num++;
          if (num>=fnum)
            num=0;
        }
      }
    }
  }
}

void main_loop (void *pvParameters)
{
  int push_btn;
  char str[16];
  int sz;

  while(1) {
    StickX = analogRead(StickXPin)/4;
    StickY = analogRead(StickYPin)/4;
    OldA=ButtonA;
    OldB=ButtonB;
    ButtonA = digitalRead(ButtonAPin);
    ButtonB = digitalRead(ButtonBPin);

    // Idle
    if (fmode==0){
      I2S_stop=0;
    
      if (StickY<256){
        num++;
        vTaskDelay(150);
      }
      if (StickY>768){
        num--;
        vTaskDelay(150);
      }
      if (num<0)
        num=fnum-1;
      if (num>=fnum)
        num=0;

      if (ButtonA==0 && OldA==1){
        fmode=1;
      }

      if (ButtonB==0 && OldB==1){
        fmode=0;
        num=0;
        old_num=-1;
      }

      if (num==0)
        disp_num=0;
      else if (num==fnum && fnum>=MAX_LINE)
        disp_num=num-MAX_LINE+1;
      else if (num-disp_num>=(MAX_LINE-1))
        disp_num=num-MAX_LINE+2;
      else if (num-disp_num<=0)
        disp_num=num-1;

      if (num!=old_num)
        show_wav_sel();

    // Play
    }else if (fmode==2){
      if (ButtonB==0 && OldB==1){
        I2S_stop=1;
      }
    }

    old_num=num;
    
    vTaskDelay(50);
  }
}

void printDirectory(File dir, int numTabs, char *typestr0, char *typestr1) {
  int i;
  while (true) {

    File entry =  dir.openNextFile();
    if (! entry) {
      // no more files
      for(i=strlen(old_fname)-2;i>=0;i--){
        if (old_fname[i]=='/'){
          old_fname[i+1]='\0';
          break;
        }
      }
      break;
    }
    // for (uint8_t i = 0; i < numTabs; i++) {
    //   Serial.print('\t');
    // }
    // Serial.println(entry.name());
    if (fnum<MAX_FILE && stristr(entry.name(), typestr0)!=NULL){
      ftype[fnum]=0;
      strcpy(fname[fnum], old_fname);
      strcat(fname[fnum++], entry.name());
    }else if (fnum<MAX_FILE && stristr(entry.name(), typestr1)!=NULL){
      ftype[fnum]=1;
      strcpy(fname[fnum], old_fname);
      strcat(fname[fnum++], entry.name());
    }else if(fnum>=MAX_FILE){
      entry.close();
      return;
    }
    
    if (entry.isDirectory()) {
      strcat(old_fname, entry.name());
      strcat(old_fname, "/");
      // Serial.println("/");
      printDirectory(entry, numTabs + 1, typestr0, typestr1);
    // } else {
    //   // files have sizes, directories do not
    //   Serial.print("\t\t");
    //   Serial.println(entry.size(), DEC);
    }
    entry.close();
  }
}

uint16_t read16(File &f) {
  uint16_t result;
  ((uint8_t *)&result)[0] = f.read(); // LSB
  ((uint8_t *)&result)[1] = f.read(); // MSB
  return result;
}

uint32_t read32(File &f) {
  uint32_t result;
  ((uint8_t *)&result)[0] = f.read(); // LSB
  ((uint8_t *)&result)[1] = f.read();
  ((uint8_t *)&result)[2] = f.read();
  ((uint8_t *)&result)[3] = f.read(); // MSB
  return result;
}

char *stristr(const char *org, char *str)
{
  char *tmp, *tmp0, *tmp1;
  tmp=(char *)org;
  for(;*tmp!='\0';tmp++){
    tmp0=tmp;
    tmp1=str;
    for(;*tmp1!='\0';tmp1++){
      if (*tmp0==*tmp1){
        tmp0++;
        continue;
      }else if ('A'<=*tmp1 && *tmp1<='Z' && *tmp0==(*tmp1-'A'+'a')){
        tmp0++;
        continue;
      }else if ('a'<=*tmp1 && *tmp1<='z' && *tmp0==(*tmp1-'a'+'A')){
        tmp0++;
        continue;
      }else
        break;
    }
    if (*tmp1=='\0')
      return(tmp);
  }
  return(NULL);
}

void show_no_sd()
{
  tft.fillRect(0, 0, _width, _height, ILI9163_BLACK);
  tft.setTextSize(1);
  tft.setTextColor(ILI9163_RED, ILI9163_BLACK);
  tft.setCursor(16, 0);
  tft.print("Error");

  tft.setTextSize(1);
  tft.setTextColor(ILI9163_WHITE, ILI9163_BLACK);
  tft.setCursor(0, 32);
  tft.print("No SD card");
  tft.display();
}

void show_wav_sel()
{
  int i;
  int j;
  char str[256];
  
  tft.fillRect(0, 0, _width, _height-9, ILI9163_BLACK);
  tft.setTextSize(1);
  tft.setTextColor(ILI9163_GREEN);
  tft.setCursor(0, 0);
  tft.print("MP3/WAV player");

  tft.setTextSize(1);
  tft.setTextColor(ILI9163_WHITE);
  if (fnum<=0){
    tft.setCursor(0, 16);
    tft.print("No data");
    return;
  }

  tft.fillRect(0, 16+(num-disp_num)*8, _width, 8, ILI9163_RED);
  for(i=0;i+disp_num<fnum && i<MAX_LINE;i++){
    j=i+disp_num;
    tft.setCursor(0, 16+i*8);
    strcpy(str, fname[j]);
    str[21]='\0';
    tft.print(str);
  }
  tft.display();
}
void show_play()
{
  tft.fillRect(0, 0, _width, _height-9, ILI9163_BLACK);
  tft.setTextSize(1);
  tft.setTextColor(ILI9163_WHITE, ILI9163_BLACK);
  tft.setCursor(0, _height/2-16);
  tft.print(fname[num]);
  tft.print("                                                                ");
  tft.display();
}
